package worker

import (
	"context"
	"errors"
	"fmt"
	"time"

	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/mongo/options"
	"go.uber.org/zap"
	"gitea.pena/PenaSide/docxTemplater/amo"
	"gitea.pena/PenaSide/docxTemplater/dal"
	"gitea.pena/PenaSide/docxTemplater/dal/model"
	"gitea.pena/PenaSide/docxTemplater/dal/mongos"
	"gitea.pena/PenaSide/docxTemplater/gdisk"
	"gitea.pena/PenaSide/docxTemplater/yadisk"
)

const (
	SuperVisorTick = 2 * time.Minute
)

type SuperVisorDeps struct {
	Dal    *dal.MongoDAL
	YaDisk *yadisk.ClientApp
	GDisk  *gdisk.ClientApp
	Amo    *amo.ClientApp
	Logger *zap.Logger
}

type SuperVisor struct {
	dal            *dal.MongoDAL
	yaDisk         *yadisk.ClientApp
	gDisk          *gdisk.ClientApp
	amo            *amo.ClientApp
	logger         *zap.Logger
	pool           *Pool
	stopWorker     chan string
	workerTaskChan chan model.WorkerTask
	interrupter    chan bool
}

func NewSuperVisor(deps SuperVisorDeps) (*SuperVisor, error) {
	if deps.Dal == nil {
		return nil, errors.New("nil SuperVisorDeps.Dal")
	}

	if deps.YaDisk == nil {
		return nil, errors.New("nil SuperVisorDeps.YaDisk")
	}

	if deps.GDisk == nil {
		return nil, errors.New("nil SuperVisorDeps.GDisk")
	}

	if deps.Amo == nil {
		return nil, errors.New("nil SuperVisorDeps.Amo")
	}

	if deps.Logger == nil {
		return nil, errors.New("nil SuperVisorDeps.Logger")
	}

	return &SuperVisor{
		dal:            deps.Dal,
		yaDisk:         deps.YaDisk,
		gDisk:          deps.GDisk,
		amo:            deps.Amo,
		logger:         deps.Logger,
		pool:           NewPool(),
		stopWorker:     make(chan string),
		workerTaskChan: make(chan model.WorkerTask),
		interrupter:    make(chan bool),
	}, nil
}

func (sv *SuperVisor) Start(ctx context.Context) {
	// Проверяем задачи со статусом new, pending и processing, если не просрочены отправляем их в работу
	tasks, err := sv.dal.WorkerTask.GetByFilter(
		ctx,
		&mongos.WorkerStatusFilter{Status: model.WorkerTaskStatusNew},
		options.Find().SetSort(bson.D{{Key: "created_at", Value: 1}}),
	)
	if err != nil {
		sv.logger.Error("GetByFilter", zap.Error(err))
	}

	if len(tasks) > 0 {
		sv.logger.Info(fmt.Sprintf("got %v tasks with status=new", len(tasks)))
	}

	for index := range tasks {
		if tasks[index].UpdatedAt.Add(TaskTimeout).Before(time.Now()) {
			sv.SetTaskTimeout(ctx, &tasks[index])
		} else {
			go sv.SetTask(ctx, tasks[index])
		}
	}

	// На всякий случай ставлю слип, чтобы воркеры успели взять в работу текущие задачи
	time.Sleep(time.Second * 5)

	tasks, err = sv.dal.WorkerTask.GetByFilter(
		ctx,
		&mongos.WorkerStatusFilter{Status: model.WorkerTaskStatusProcessing},
		options.Find().SetSort(bson.D{{Key: "created_at", Value: 1}}),
	)
	if err != nil {
		sv.logger.Error("GetByFilter", zap.Error(err))
	}

	if len(tasks) > 0 {
		sv.logger.Info(fmt.Sprintf("got %v tasks with status=processing", len(tasks)))
	}

	for index := range tasks {
		if tasks[index].UpdatedAt.Add(TaskTimeout).Before(time.Now()) {
			sv.SetTaskTimeout(ctx, &tasks[index])
		} else {
			go sv.SetTask(ctx, tasks[index])
		}
	}

	// На всякий случай ставлю слип, чтобы воркеры успели взять в работу текущие задачи
	time.Sleep(time.Second * 5)

	tasks, err = sv.dal.WorkerTask.GetByFilter(
		ctx,
		&mongos.WorkerStatusFilter{Status: model.WorkerTaskStatusPending},
		options.Find().SetSort(bson.D{{Key: "created_at", Value: 1}}),
	)
	if err != nil {
		sv.logger.Error("GetByFilter", zap.Error(err))
	}

	if len(tasks) > 0 {
		sv.logger.Info(fmt.Sprintf("got %v tasks with status=pending", len(tasks)))
	}

	for index := range tasks {
		if tasks[index].UpdatedAt.Add(TaskTimeout).Before(time.Now()) {
			sv.SetTaskTimeout(ctx, &tasks[index])
		} else {
			go sv.SetTask(ctx, tasks[index])
		}
	}

	// На всякий случай ставлю слип, чтобы воркеры успели взять в работу текущие задачи
	time.Sleep(time.Second * 5)

	// Запускаем горутину c получением новых задач
	go func() {
		workerTaskChan := make(chan model.WorkerTask)
		go sv.dal.WorkerTask.Listen(ctx, model.WorkerTaskStatusNew, workerTaskChan)

		for task := range workerTaskChan {
			go sv.SetTask(ctx, task)
		}
	}()

	// Запускаем горутину проверяющую просрочку задач по статусу и отвечающую за закрытие супервизора
	go func() {
		ticker := time.NewTicker(SuperVisorTick)
		defer ticker.Stop()

		for {
			// Проверяем задачи в processing и при просрочке ставим статус timeout
			tasks, err = sv.dal.WorkerTask.GetByFilter(
				ctx,
				&mongos.WorkerStatusFilter{Status: model.WorkerTaskStatusProcessing},
				options.Find().SetSort(bson.D{{Key: "created_at", Value: 1}}),
			)
			if err != nil {
				sv.logger.Error("GetByFilter", zap.Error(err))
			}

			if len(tasks) > 0 {
				sv.logger.Info(fmt.Sprintf("got %v tasks with status=processing", len(tasks)))
			}

			for index := range tasks {
				if tasks[index].UpdatedAt.Add(TaskTimeout).Before(time.Now()) {
					sv.SetTaskTimeout(ctx, &tasks[index])
				}
			}

			// Проверяем задачи в pending и при просрочке ставим статус timeout
			tasks, err = sv.dal.WorkerTask.GetByFilter(
				ctx,
				&mongos.WorkerStatusFilter{Status: model.WorkerTaskStatusPending},
				options.Find().SetSort(bson.D{{Key: "created_at", Value: 1}}),
			)
			if err != nil {
				sv.logger.Error("GetByFilter", zap.Error(err))
			}

			if len(tasks) > 0 {
				sv.logger.Info(fmt.Sprintf("got %v tasks with status=processing", len(tasks)))
			}

			for i := range tasks {
				if tasks[i].UpdatedAt.Add(TaskTimeout).Before(time.Now()) {
					sv.SetTaskTimeout(ctx, &tasks[i])
				}
			}

			select {
			case <-ctx.Done():
				sv.Stop()
				return
			case <-ticker.C:
				continue
			case <-sv.interrupter:
				return
			}
		}
	}()

	// Запускаем горутину закрывающие воркеры по lifeTimer
	go func() {
		for {
			select {
			case workerAmoID := <-sv.stopWorker:
				sv.pool.Unset(workerAmoID)
				continue
			case <-ctx.Done():
				sv.Stop()
				return
			case <-sv.interrupter:
				return
			}
		}
	}()
}

// SetTask - ставит задаче статус pending, отправляет её в воркер. Если воркер не существует, то создает его.
func (sv *SuperVisor) SetTask(ctx context.Context, task model.WorkerTask) {
	var err error

	// Ставим статус ожидания
	if err = sv.dal.WorkerTask.UpdateStatus(ctx, task.ID, model.WorkerTaskStatusPending); err != nil {
		sv.logger.Error("cannot update workerTask status", zap.Error(err), zap.String("_id", task.ID))
	}

	// Если у задачи отсутствует amoId ставим статус failed и пишем в историю
	if task.AmoID == "" {
		err = errors.New("task have no amo_id")
		sv.logger.Error(err.Error(), zap.String("_id", task.ID))

		var historyFound []model.History
		historyFound, err = sv.dal.History.GetByFilter(ctx, &mongos.HistoryFilter{WorkerTaskID: task.ID})

		if err != nil {
			sv.logger.Error("cannot get history", zap.Error(err), zap.String("_id", task.ID))
		} else if len(historyFound) > 0 {
			historyFound[0].Errors = append(historyFound[0].Errors, err.Error())
			err = sv.dal.History.UpdateByID(ctx, &historyFound[0])

			if err != nil {
				sv.logger.Error("cannot update history", zap.Error(err), zap.String("_id", task.ID))
			}
		}

		err = sv.dal.WorkerTask.UpdateStatus(ctx, task.ID, model.WorkerTaskStatusFailed)
		if err != nil {
			sv.logger.Error("cannot update workerTask status", zap.Error(err), zap.String("_id", task.ID))
		}
		return
	}

	// Создаем воркеры
	worker := sv.pool.Get(task.AmoID)
	if worker == nil {
		workerDeps := WorkerDeps{
			AmoID:  task.AmoID,
			Dal:    sv.dal,
			YaDisk: sv.yaDisk,
			GDisk:  sv.gDisk,
			Amo:    sv.amo,
			Logger: sv.logger,
		}

		worker, err = NewWorker(workerDeps)
		if err != nil {
			sv.logger.Error("FailedCreateWorker", zap.Error(err))
			return
		}

		sv.pool.Set(task.AmoID, worker)
		worker.Start(ctx, sv.stopWorker)
	}

	sv.pool.Get(task.AmoID).Tube <- &task
}

func (sv *SuperVisor) SetTaskTimeout(ctx context.Context, task *model.WorkerTask) {
	sv.logger.Info("task timeout", zap.String("_id", task.ID))
	historyFound, err := sv.dal.History.GetByFilter(ctx, &mongos.HistoryFilter{WorkerTaskID: task.ID})
	if err != nil {
		sv.logger.Error("cannot get history", zap.Error(err), zap.String("_id", task.ID))
	}

	if len(historyFound) > 0 {
		err = errors.New("task timeout")
		if len(historyFound[0].Errors) > 0 {
			historyFound[0].Errors = append(historyFound[0].Errors, err.Error())
		} else {
			historyFound[0].Errors = []string{err.Error()}
		}

		if err = sv.dal.History.UpdateByID(ctx, &historyFound[0]); err != nil {
			sv.logger.Error("cannot update history", zap.Error(err), zap.String("_id", task.ID))
		}
	}

	if err = sv.dal.WorkerTask.UpdateStatus(ctx, task.ID, model.WorkerTaskStatusTimeout); err != nil {
		sv.logger.Error("cannot update workerTask status", zap.Error(err), zap.String("_id", task.ID))
	}
}

func (sv *SuperVisor) Stop() {
	for _, worker := range sv.pool.workers {
		sv.pool.Unset(worker.AmoID)
	}
	close(sv.workerTaskChan)
	sv.interrupter <- true
}
